@@ -1,4 +1,4 @@
-from importlib.metadata import version
+from importlib.metadata import version as _metadata_version
 
 from .agent import Agent, CallToolsNode, EndStrategy, ModelRequestNode, UserPromptNode, capture_run_messages
 from .exceptions import (
@@ -10,7 +10,9 @@ from .exceptions import (
     UsageLimitExceeded,
     UserError,
 )
-from .messages import AudioUrl, BinaryContent, DocumentUrl, ImageUrl
+from .format_prompt import format_as_xml
+from .messages import AudioUrl, BinaryContent, DocumentUrl, ImageUrl, VideoUrl
+from .result import ToolOutput
 from .tools import RunContext, Tool
 
 __all__ = (
@@ -33,10 +35,15 @@ __all__ = (
     # messages
     'ImageUrl',
     'AudioUrl',
+    'VideoUrl',
     'DocumentUrl',
     'BinaryContent',
     # tools
     'Tool',
     'RunContext',
+    # result
+    'ToolOutput',
+    # format_prompt
+    'format_as_xml',
 )
-__version__ = version('pydantic_ai_slim')
+__version__ = _metadata_version('pydantic_ai_slim')
