@@ -2,21 +2,26 @@ from __future__ import annotations as _annotations
 
 import argparse
 import asyncio
+import importlib
+import os
 import sys
 from asyncio import CancelledError
 from collections.abc import Sequence
 from contextlib import ExitStack
 from datetime import datetime, timezone
-from importlib.metadata import version
 from pathlib import Path
 from typing import Any, cast
 
 from typing_inspection.introspection import get_literal_values
 
-from pydantic_ai.agent import Agent
-from pydantic_ai.exceptions import UserError
-from pydantic_ai.messages import ModelMessage, PartDeltaEvent, TextPartDelta
-from pydantic_ai.models import KnownModelName, infer_model
+from pydantic_ai.result import OutputDataT
+from pydantic_ai.tools import AgentDepsT
+
+from . import __version__
+from .agent import Agent
+from .exceptions import UserError
+from .messages import ModelMessage
+from .models import KnownModelName, infer_model
 
 try:
     import argcomplete
@@ -39,11 +44,20 @@ except ImportError as _import_error:
     ) from _import_error
 
 
-__version__ = version('pydantic-ai-slim')
+__all__ = 'cli', 'cli_exit'
+
+
+PYDANTIC_AI_HOME = Path.home() / '.pydantic-ai'
+"""The home directory for PydanticAI CLI.
+
+This folder is used to store the prompt history and configuration.
+"""
+
+PROMPT_HISTORY_FILENAME = 'prompt-history.txt'
 
 
 class SimpleCodeBlock(CodeBlock):
-    """Customised code blocks in markdown.
+    """Customized code blocks in markdown.
 
     This avoids a background color which messes up copy-pasting and sets the language name as dim prefix and suffix.
     """
@@ -56,7 +70,7 @@ class SimpleCodeBlock(CodeBlock):
 
 
 class LeftHeading(Heading):
-    """Customised headings in markdown to stop centering and prepend markdown style hashes."""
+    """Customized headings in markdown to stop centering and prepend markdown style hashes."""
 
     def __rich_console__(self, console: Console, options: ConsoleOptions) -> RenderResult:
         # note we use `Style(bold=True)` not `self.style_name` here to disable underlining which is ugly IMHO
@@ -83,14 +97,20 @@ The current date and time is {datetime.now()} {tzname}.
 The user is running {sys.platform}."""
 
 
-def cli(args_list: Sequence[str] | None = None) -> int:
+def cli_exit(prog_name: str = 'pai'):  # pragma: no cover
+    """Run the CLI and exit."""
+    sys.exit(cli(prog_name=prog_name))
+
+
+def cli(args_list: Sequence[str] | None = None, *, prog_name: str = 'pai') -> int:  # noqa: C901
+    """Run the CLI and return the exit code for the process."""
     parser = argparse.ArgumentParser(
-        prog='pai',
+        prog=prog_name,
         description=f"""\
 PydanticAI CLI v{__version__}\n\n
 
-Special prompt:
-* `/exit` - exit the interactive mode
+Special prompts:
+* `/exit` - exit the interactive mode (ctrl-c and ctrl-d also work)
 * `/markdown` - show the last markdown output of the last question
 * `/multiline` - toggle multiline mode
 """,
@@ -101,13 +121,17 @@ Special prompt:
         '-m',
         '--model',
         nargs='?',
-        help='Model to use, in format "<provider>:<model>" e.g. "openai:gpt-4o". Defaults to "openai:gpt-4o".',
-        default='openai:gpt-4o',
+        help='Model to use, in format "<provider>:<model>" e.g. "openai:gpt-4o" or "anthropic:claude-3-7-sonnet-latest". Defaults to "openai:gpt-4o".',
     )
     # we don't want to autocomplete or list models that don't include the provider,
     # e.g. we want to show `openai:gpt-4o` but not `gpt-4o`
     qualified_model_names = [n for n in get_literal_values(KnownModelName.__value__) if ':' in n]
     arg.completer = argcomplete.ChoicesCompleter(qualified_model_names)  # type: ignore[reportPrivateUsage]
+    parser.add_argument(
+        '-a',
+        '--agent',
+        help='Custom Agent to use, in format "module:variable", e.g. "mymodule.submodule:my_agent"',
+    )
     parser.add_argument(
         '-l',
         '--list-models',
@@ -118,32 +142,59 @@ Special prompt:
         '-t',
         '--code-theme',
         nargs='?',
-        help='Which colors to use for code, can be "dark", "light" or any theme from pygments.org/styles/. Defaults to "monokai".',
-        default='monokai',
+        help='Which colors to use for code, can be "dark", "light" or any theme from pygments.org/styles/. Defaults to "dark" which works well on dark terminals.',
+        default='dark',
     )
-    parser.add_argument('--no-stream', action='store_true', help='Whether to stream responses from the model')
+    parser.add_argument('--no-stream', action='store_true', help='Disable streaming from the model')
     parser.add_argument('--version', action='store_true', help='Show version and exit')
 
     argcomplete.autocomplete(parser)
     args = parser.parse_args(args_list)
 
     console = Console()
-    console.print(
-        f'[green]pai - PydanticAI CLI v{__version__} using[/green] [magenta]{args.model}[/magenta]', highlight=False
-    )
+    name_version = f'[green]{prog_name} - PydanticAI CLI v{__version__}[/green]'
     if args.version:
+        console.print(name_version, highlight=False)
         return 0
     if args.list_models:
-        console.print('Available models:', style='green bold')
+        console.print(f'{name_version}\n\n[green]Available models:[/green]')
         for model in qualified_model_names:
             console.print(f'  {model}', highlight=False)
         return 0
 
-    try:
-        cli_agent.model = infer_model(args.model)
-    except UserError as e:
-        console.print(f'Error initializing [magenta]{args.model}[/magenta]:\n[red]{e}[/red]')
-        return 1
+    agent: Agent[None, str] = cli_agent
+    if args.agent:
+        sys.path.append(os.getcwd())
+        try:
+            module_path, variable_name = args.agent.split(':')
+        except ValueError:
+            console.print('[red]Error: Agent must be specified in "module:variable" format[/red]')
+            return 1
+
+        module = importlib.import_module(module_path)
+        agent = getattr(module, variable_name)
+        if not isinstance(agent, Agent):
+            console.print(f'[red]Error: {args.agent} is not an Agent instance[/red]')
+            return 1
+
+    model_arg_set = args.model is not None
+    if agent.model is None or model_arg_set:
+        try:
+            agent.model = infer_model(args.model or 'openai:gpt-4o')
+        except UserError as e:
+            console.print(f'Error initializing [magenta]{args.model}[/magenta]:\n[red]{e}[/red]')
+            return 1
+
+    model_name = agent.model if isinstance(agent.model, str) else f'{agent.model.system}:{agent.model.model_name}'
+    if args.agent and model_arg_set:
+        console.print(
+            f'{name_version} using custom agent [magenta]{args.agent}[/magenta] with [magenta]{model_name}[/magenta]',
+            highlight=False,
+        )
+    elif args.agent:
+        console.print(f'{name_version} using custom agent [magenta]{args.agent}[/magenta]', highlight=False)
+    else:
+        console.print(f'{name_version} with [magenta]{model_name}[/magenta]', highlight=False)
 
     stream = not args.no_stream
     if args.code_theme == 'light':
@@ -151,32 +202,42 @@ Special prompt:
     elif args.code_theme == 'dark':
         code_theme = 'monokai'
     else:
-        code_theme = args.code_theme
+        code_theme = args.code_theme  # pragma: no cover
 
     if prompt := cast(str, args.prompt):
         try:
-            asyncio.run(ask_agent(cli_agent, prompt, stream, console, code_theme))
+            asyncio.run(ask_agent(agent, prompt, stream, console, code_theme))
         except KeyboardInterrupt:
             pass
         return 0
 
-    history = Path.home() / '.pai-prompt-history.txt'
-    # doing this instead of `PromptSession[Any](history=` allows mocking of PromptSession in tests
-    session: PromptSession[Any] = PromptSession(history=FileHistory(str(history)))
     try:
-        return asyncio.run(run_chat(session, stream, cli_agent, console, code_theme))
+        return asyncio.run(run_chat(stream, agent, console, code_theme, prog_name))
     except KeyboardInterrupt:  # pragma: no cover
         return 0
 
 
-async def run_chat(session: PromptSession[Any], stream: bool, agent: Agent, console: Console, code_theme: str) -> int:
+async def run_chat(
+    stream: bool,
+    agent: Agent[AgentDepsT, OutputDataT],
+    console: Console,
+    code_theme: str,
+    prog_name: str,
+    config_dir: Path | None = None,
+    deps: AgentDepsT = None,
+) -> int:
+    prompt_history_path = (config_dir or PYDANTIC_AI_HOME) / PROMPT_HISTORY_FILENAME
+    prompt_history_path.parent.mkdir(parents=True, exist_ok=True)
+    prompt_history_path.touch(exist_ok=True)
+    session: PromptSession[Any] = PromptSession(history=FileHistory(str(prompt_history_path)))
+
     multiline = False
     messages: list[ModelMessage] = []
 
     while True:
         try:
             auto_suggest = CustomAutoSuggest(['/markdown', '/multiline', '/exit'])
-            text = await session.prompt_async('pai ➤ ', auto_suggest=auto_suggest, multiline=multiline)
+            text = await session.prompt_async(f'{prog_name} ➤ ', auto_suggest=auto_suggest, multiline=multiline)
         except (KeyboardInterrupt, EOFError):  # pragma: no cover
             return 0
 
@@ -190,42 +251,40 @@ async def run_chat(session: PromptSession[Any], stream: bool, agent: Agent, cons
                 return exit_value
         else:
             try:
-                messages = await ask_agent(agent, text, stream, console, code_theme, messages)
+                messages = await ask_agent(agent, text, stream, console, code_theme, deps, messages)
             except CancelledError:  # pragma: no cover
                 console.print('[dim]Interrupted[/dim]')
 
 
 async def ask_agent(
-    agent: Agent,
+    agent: Agent[AgentDepsT, OutputDataT],
     prompt: str,
     stream: bool,
     console: Console,
     code_theme: str,
+    deps: AgentDepsT = None,
     messages: list[ModelMessage] | None = None,
 ) -> list[ModelMessage]:
     status = Status('[dim]Working on it…[/dim]', console=console)
 
     if not stream:
         with status:
-            result = await agent.run(prompt, message_history=messages)
-        content = result.data
+            result = await agent.run(prompt, message_history=messages, deps=deps)
+        content = str(result.output)
         console.print(Markdown(content, code_theme=code_theme))
         return result.all_messages()
 
     with status, ExitStack() as stack:
-        async with agent.iter(prompt, message_history=messages) as agent_run:
-            live = Live('', refresh_per_second=15, console=console, vertical_overflow='visible')
-            content: str = ''
+        async with agent.iter(prompt, message_history=messages, deps=deps) as agent_run:
+            live = Live('', refresh_per_second=15, console=console, vertical_overflow='ellipsis')
             async for node in agent_run:
                 if Agent.is_model_request_node(node):
                     async with node.stream(agent_run.ctx) as handle_stream:
                         status.stop()  # stopping multiple times is idempotent
                         stack.enter_context(live)  # entering multiple times is idempotent
 
-                        async for event in handle_stream:
-                            if isinstance(event, PartDeltaEvent) and isinstance(event.delta, TextPartDelta):
-                                content += event.delta.content_delta
-                                live.update(Markdown(content, code_theme=code_theme))
+                        async for content in handle_stream.stream_output(debounce_by=None):
+                            live.update(Markdown(str(content), code_theme=code_theme))
 
         assert agent_run.result is not None
         return agent_run.result.all_messages()
@@ -285,7 +344,3 @@ def handle_slash_command(
     else:
         console.print(f'[red]Unknown command[/red] [magenta]`{ident_prompt}`[/magenta]')
     return None, multiline
-
-
-def app():  # pragma: no cover
-    sys.exit(cli())
