@@ -58,7 +58,7 @@ def function_schema(  # noqa: C901
     Returns:
         A `FunctionSchema` instance.
     """
-    config = ConfigDict(title=function.__name__)
+    config = ConfigDict(title=function.__name__, use_attribute_docstrings=True)
     config_wrapper = ConfigWrapper(config)
     gen_schema = _generate_schema.GenerateSchema(config_wrapper)
 
@@ -76,8 +76,15 @@ def function_schema(  # noqa: C901
     description, field_descriptions = doc_descriptions(function, sig, docstring_format=docstring_format)
 
     if require_parameter_descriptions:
-        if len(field_descriptions) != len(sig.parameters):
+        if takes_ctx:
+            parameters_without_ctx = set(
+                name for name in sig.parameters if not _is_call_ctx(sig.parameters[name].annotation)
+            )
+            missing_params = parameters_without_ctx - set(field_descriptions)
+        else:
             missing_params = set(sig.parameters) - set(field_descriptions)
+
+        if missing_params:
             errors.append(f'Missing parameter descriptions for {", ".join(missing_params)}')
 
     for index, (name, p) in enumerate(sig.parameters.items()):
