@@ -12,30 +12,39 @@ from contextlib import asynccontextmanager, contextmanager
 from dataclasses import dataclass, field
 from datetime import datetime
 from functools import cache
-from typing import TYPE_CHECKING
 
 import httpx
 from typing_extensions import Literal, TypeAliasType
 
 from .._parts_manager import ModelResponsePartsManager
 from ..exceptions import UserError
-from ..messages import ModelMessage, ModelResponse, ModelResponseStreamEvent
+from ..messages import ModelMessage, ModelRequest, ModelResponse, ModelResponseStreamEvent
 from ..settings import ModelSettings
+from ..tools import ToolDefinition
 from ..usage import Usage
 
-if TYPE_CHECKING:
-    from ..tools import ToolDefinition
-
-
 KnownModelName = TypeAliasType(
     'KnownModelName',
     Literal[
-        'anthropic:claude-3-7-sonnet-latest',
+        'anthropic:claude-2.0',
+        'anthropic:claude-2.1',
+        'anthropic:claude-3-5-haiku-20241022',
         'anthropic:claude-3-5-haiku-latest',
+        'anthropic:claude-3-5-sonnet-20240620',
+        'anthropic:claude-3-5-sonnet-20241022',
         'anthropic:claude-3-5-sonnet-latest',
+        'anthropic:claude-3-7-sonnet-20250219',
+        'anthropic:claude-3-7-sonnet-latest',
+        'anthropic:claude-3-haiku-20240307',
+        'anthropic:claude-3-opus-20240229',
         'anthropic:claude-3-opus-latest',
-        'claude-3-7-sonnet-latest',
-        'claude-3-5-haiku-latest',
+        'anthropic:claude-3-sonnet-20240229',
+        'anthropic:claude-4-opus-20250514',
+        'anthropic:claude-4-sonnet-20250514',
+        'anthropic:claude-opus-4-0',
+        'anthropic:claude-opus-4-20250514',
+        'anthropic:claude-sonnet-4-0',
+        'anthropic:claude-sonnet-4-20250514',
         'bedrock:amazon.titan-tg1-large',
         'bedrock:amazon.titan-text-lite-v1',
         'bedrock:amazon.titan-text-express-v1',
@@ -79,8 +88,25 @@ KnownModelName = TypeAliasType(
         'bedrock:mistral.mixtral-8x7b-instruct-v0:1',
         'bedrock:mistral.mistral-large-2402-v1:0',
         'bedrock:mistral.mistral-large-2407-v1:0',
+        'claude-2.0',
+        'claude-2.1',
+        'claude-3-5-haiku-20241022',
+        'claude-3-5-haiku-latest',
+        'claude-3-5-sonnet-20240620',
+        'claude-3-5-sonnet-20241022',
         'claude-3-5-sonnet-latest',
+        'claude-3-7-sonnet-20250219',
+        'claude-3-7-sonnet-latest',
+        'claude-3-haiku-20240307',
+        'claude-3-opus-20240229',
         'claude-3-opus-latest',
+        'claude-3-sonnet-20240229',
+        'claude-4-opus-20250514',
+        'claude-4-sonnet-20250514',
+        'claude-opus-4-0',
+        'claude-opus-4-20250514',
+        'claude-sonnet-4-0',
+        'claude-sonnet-4-20250514',
         'cohere:c4ai-aya-expanse-32b',
         'cohere:c4ai-aya-expanse-8b',
         'cohere:command',
@@ -96,28 +122,26 @@ KnownModelName = TypeAliasType(
         'cohere:command-r7b-12-2024',
         'deepseek:deepseek-chat',
         'deepseek:deepseek-reasoner',
-        'google-gla:gemini-1.0-pro',
         'google-gla:gemini-1.5-flash',
         'google-gla:gemini-1.5-flash-8b',
         'google-gla:gemini-1.5-pro',
-        'google-gla:gemini-2.0-flash-exp',
-        'google-gla:gemini-2.0-flash-thinking-exp-01-21',
-        'google-gla:gemini-exp-1206',
+        'google-gla:gemini-1.0-pro',
         'google-gla:gemini-2.0-flash',
         'google-gla:gemini-2.0-flash-lite-preview-02-05',
         'google-gla:gemini-2.0-pro-exp-02-05',
+        'google-gla:gemini-2.5-flash-preview-05-20',
         'google-gla:gemini-2.5-pro-exp-03-25',
-        'google-vertex:gemini-1.0-pro',
+        'google-gla:gemini-2.5-pro-preview-05-06',
         'google-vertex:gemini-1.5-flash',
         'google-vertex:gemini-1.5-flash-8b',
         'google-vertex:gemini-1.5-pro',
-        'google-vertex:gemini-2.0-flash-exp',
-        'google-vertex:gemini-2.0-flash-thinking-exp-01-21',
-        'google-vertex:gemini-exp-1206',
+        'google-vertex:gemini-1.0-pro',
         'google-vertex:gemini-2.0-flash',
         'google-vertex:gemini-2.0-flash-lite-preview-02-05',
         'google-vertex:gemini-2.0-pro-exp-02-05',
+        'google-vertex:gemini-2.5-flash-preview-05-20',
         'google-vertex:gemini-2.5-pro-exp-03-25',
+        'google-vertex:gemini-2.5-pro-preview-05-06',
         'gpt-3.5-turbo',
         'gpt-3.5-turbo-0125',
         'gpt-3.5-turbo-0301',
@@ -137,6 +161,12 @@ KnownModelName = TypeAliasType(
         'gpt-4-turbo-2024-04-09',
         'gpt-4-turbo-preview',
         'gpt-4-vision-preview',
+        'gpt-4.1',
+        'gpt-4.1-2025-04-14',
+        'gpt-4.1-mini',
+        'gpt-4.1-mini-2025-04-14',
+        'gpt-4.1-nano',
+        'gpt-4.1-nano-2025-04-14',
         'gpt-4o',
         'gpt-4o-2024-05-13',
         'gpt-4o-2024-08-06',
@@ -184,6 +214,8 @@ KnownModelName = TypeAliasType(
         'o1-mini-2024-09-12',
         'o1-preview',
         'o1-preview-2024-09-12',
+        'o3',
+        'o3-2025-04-16',
         'o3-mini',
         'o3-mini-2025-01-31',
         'openai:chatgpt-4o-latest',
@@ -206,6 +238,12 @@ KnownModelName = TypeAliasType(
         'openai:gpt-4-turbo-2024-04-09',
         'openai:gpt-4-turbo-preview',
         'openai:gpt-4-vision-preview',
+        'openai:gpt-4.1',
+        'openai:gpt-4.1-2025-04-14',
+        'openai:gpt-4.1-mini',
+        'openai:gpt-4.1-mini-2025-04-14',
+        'openai:gpt-4.1-nano',
+        'openai:gpt-4.1-nano-2025-04-14',
         'openai:gpt-4o',
         'openai:gpt-4o-2024-05-13',
         'openai:gpt-4o-2024-08-06',
@@ -227,8 +265,12 @@ KnownModelName = TypeAliasType(
         'openai:o1-mini-2024-09-12',
         'openai:o1-preview',
         'openai:o1-preview-2024-09-12',
+        'openai:o3',
+        'openai:o3-2025-04-16',
         'openai:o3-mini',
         'openai:o3-mini-2025-01-31',
+        'openai:o4-mini',
+        'openai:o4-mini-2025-04-16',
         'test',
     ],
 )
@@ -240,11 +282,11 @@ KnownModelName = TypeAliasType(
 
 @dataclass
 class ModelRequestParameters:
-    """Configuration for an agent's request to a model, specifically related to tools and result handling."""
+    """Configuration for an agent's request to a model, specifically related to tools and output handling."""
 
-    function_tools: list[ToolDefinition]
-    allow_text_result: bool
-    result_tools: list[ToolDefinition]
+    function_tools: list[ToolDefinition] = field(default_factory=list)
+    allow_text_output: bool = True
+    output_tools: list[ToolDefinition] = field(default_factory=list)
 
 
 class Model(ABC):
@@ -256,7 +298,7 @@ class Model(ABC):
         messages: list[ModelMessage],
         model_settings: ModelSettings | None,
         model_request_parameters: ModelRequestParameters,
-    ) -> tuple[ModelResponse, Usage]:
+    ) -> ModelResponse:
         """Make a request to the model."""
         raise NotImplementedError()
 
@@ -274,6 +316,15 @@ class Model(ABC):
         # noinspection PyUnreachableCode
         yield  # pragma: no cover
 
+    def customize_request_parameters(self, model_request_parameters: ModelRequestParameters) -> ModelRequestParameters:
+        """Customize the request parameters for the model.
+
+        This method can be overridden by subclasses to modify the request parameters before sending them to the model.
+        In particular, this method can be used to make modifications to the generated tool JSON schemas if necessary
+        for vendor/model-specific reasons.
+        """
+        return model_request_parameters
+
     @property
     @abstractmethod
     def model_name(self) -> str:
@@ -297,6 +348,49 @@ class Model(ABC):
         """The base URL for the provider API, if available."""
         return None
 
+    @staticmethod
+    def _get_instructions(messages: list[ModelMessage]) -> str | None:
+        """Get instructions from the first ModelRequest found when iterating messages in reverse.
+
+        In the case that a "mock" request was generated to include a tool-return part for a result tool,
+        we want to use the instructions from the second-to-most-recent request (which should correspond to the
+        original request that generated the response that resulted in the tool-return part).
+        """
+        last_two_requests: list[ModelRequest] = []
+        for message in reversed(messages):
+            if isinstance(message, ModelRequest):
+                last_two_requests.append(message)
+                if len(last_two_requests) == 2:
+                    break
+                if message.instructions is not None:
+                    return message.instructions
+
+        # If we don't have two requests, and we didn't already return instructions, there are definitely not any:
+        if len(last_two_requests) != 2:
+            return None
+
+        most_recent_request = last_two_requests[0]
+        second_most_recent_request = last_two_requests[1]
+
+        # If we've gotten this far and the most recent request consists of only tool-return parts or retry-prompt parts,
+        # we use the instructions from the second-to-most-recent request. This is necessary because when handling
+        # result tools, we generate a "mock" ModelRequest with a tool-return part for it, and that ModelRequest will not
+        # have the relevant instructions from the agent.
+
+        # While it's possible that you could have a message history where the most recent request has only tool returns,
+        # I believe there is no way to achieve that would _change_ the instructions without manually crafting the most
+        # recent message. That might make sense in principle for some usage pattern, but it's enough of an edge case
+        # that I think it's not worth worrying about, since you can work around this by inserting another ModelRequest
+        # with no parts at all immediately before the request that has the tool calls (that works because we only look
+        # at the two most recent ModelRequests here).
+
+        # If you have a use case where this causes pain, please open a GitHub issue and we can discuss alternatives.
+
+        if all(p.part_kind == 'tool-return' or p.part_kind == 'retry-prompt' for p in most_recent_request.parts):
+            return second_most_recent_request.instructions
+
+        return None
+
 
 @dataclass
 class StreamedResponse(ABC):
@@ -328,7 +422,10 @@ class StreamedResponse(ABC):
     def get(self) -> ModelResponse:
         """Build a [`ModelResponse`][pydantic_ai.messages.ModelResponse] from the data received from the stream so far."""
         return ModelResponse(
-            parts=self._parts_manager.get_parts(), model_name=self.model_name, timestamp=self.timestamp
+            parts=self._parts_manager.get_parts(),
+            model_name=self.model_name,
+            timestamp=self.timestamp,
+            usage=self.usage(),
         )
 
     def usage(self) -> Usage:
@@ -412,13 +509,13 @@ def infer_model(model: Model | KnownModelName | str) -> Model:
             raise UserError(f'Unknown model: {model}')
 
     if provider == 'vertexai':
-        provider = 'google-vertex'
+        provider = 'google-vertex'  # pragma: no cover
 
     if provider == 'cohere':
         from .cohere import CohereModel
 
         return CohereModel(model_name, provider=provider)
-    elif provider in ('deepseek', 'openai'):
+    elif provider in ('deepseek', 'openai', 'azure', 'openrouter'):
         from .openai import OpenAIModel
 
         return OpenAIModel(model_name, provider=provider)
@@ -443,7 +540,7 @@ def infer_model(model: Model | KnownModelName | str) -> Model:
 
         return BedrockConverseModel(model_name, provider=provider)
     else:
-        raise UserError(f'Unknown model: {model}')
+        raise UserError(f'Unknown model: {model}')  # pragma: no cover
 
 
 def cached_async_http_client(*, provider: str | None = None, timeout: int = 600, connect: int = 5) -> httpx.AsyncClient:
