@@ -2,17 +2,18 @@ from __future__ import annotations
 
 import functools
 import typing
-from collections.abc import AsyncIterator, Iterable
+from collections.abc import AsyncIterator, Iterable, Iterator, Mapping
 from contextlib import asynccontextmanager
 from dataclasses import dataclass, field
 from datetime import datetime
-from typing import TYPE_CHECKING, Generic, Literal, Union, cast, overload
+from itertools import count
+from typing import TYPE_CHECKING, Any, Generic, Literal, Union, cast, overload
 
 import anyio
 import anyio.to_thread
 from typing_extensions import ParamSpec, assert_never
 
-from pydantic_ai import _utils, result
+from pydantic_ai import _utils, usage
 from pydantic_ai.messages import (
     AudioUrl,
     BinaryContent,
@@ -29,6 +30,7 @@ from pydantic_ai.messages import (
     ToolCallPart,
     ToolReturnPart,
     UserPromptPart,
+    VideoUrl,
 )
 from pydantic_ai.models import Model, ModelRequestParameters, StreamedResponse, cached_async_http_client
 from pydantic_ai.providers import Provider, infer_provider
@@ -42,14 +44,20 @@ if TYPE_CHECKING:
     from mypy_boto3_bedrock_runtime.type_defs import (
         ContentBlockOutputTypeDef,
         ContentBlockUnionTypeDef,
+        ConverseRequestTypeDef,
         ConverseResponseTypeDef,
         ConverseStreamMetadataEventTypeDef,
         ConverseStreamOutputTypeDef,
+        GuardrailConfigurationTypeDef,
         ImageBlockTypeDef,
         InferenceConfigurationTypeDef,
         MessageUnionTypeDef,
+        PerformanceConfigurationTypeDef,
+        PromptVariableValuesTypeDef,
+        SystemContentBlockTypeDef,
         ToolChoiceTypeDef,
         ToolTypeDef,
+        VideoBlockTypeDef,
     )
 
 
@@ -112,10 +120,49 @@ P = ParamSpec('P')
 T = typing.TypeVar('T')
 
 
-class BedrockModelSettings(ModelSettings):
+class BedrockModelSettings(ModelSettings, total=False):
     """Settings for Bedrock models.
 
     ALL FIELDS MUST BE `bedrock_` PREFIXED SO YOU CAN MERGE THEM WITH OTHER MODELS.
+
+    See [the Bedrock Converse API docs](https://docs.aws.amazon.com/bedrock/latest/APIReference/API_runtime_Converse.html#API_runtime_Converse_RequestSyntax) for a full list.
+    See [the boto3 implementation](https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/bedrock-runtime/client/converse.html) of the Bedrock Converse API.
+    """
+
+    bedrock_guardrail_config: GuardrailConfigurationTypeDef
+    """Content moderation and safety settings for Bedrock API requests.
+
+    See more about it on <https://docs.aws.amazon.com/bedrock/latest/APIReference/API_runtime_GuardrailConfiguration.html>.
+    """
+
+    bedrock_performance_configuration: PerformanceConfigurationTypeDef
+    """Performance optimization settings for model inference.
+
+    See more about it on <https://docs.aws.amazon.com/bedrock/latest/APIReference/API_runtime_PerformanceConfiguration.html>.
+    """
+
+    bedrock_request_metadata: dict[str, str]
+    """Additional metadata to attach to Bedrock API requests.
+
+    See more about it on <https://docs.aws.amazon.com/bedrock/latest/APIReference/API_runtime_Converse.html#API_runtime_Converse_RequestSyntax>.
+    """
+
+    bedrock_additional_model_response_fields_paths: list[str]
+    """JSON paths to extract additional fields from model responses.
+
+    See more about it on <https://docs.aws.amazon.com/bedrock/latest/userguide/model-parameters.html>.
+    """
+
+    bedrock_prompt_variables: Mapping[str, PromptVariableValuesTypeDef]
+    """Variables for substitution into prompt templates.
+
+    See more about it on <https://docs.aws.amazon.com/bedrock/latest/APIReference/API_runtime_PromptVariableValues.html>.
+    """
+
+    bedrock_additional_model_requests_fields: Mapping[str, Any]
+    """Additional model-specific parameters to include in requests.
+
+    See more about it on <https://docs.aws.amazon.com/bedrock/latest/userguide/model-parameters.html>.
     """
 
 
@@ -162,8 +209,8 @@ class BedrockConverseModel(Model):
 
     def _get_tools(self, model_request_parameters: ModelRequestParameters) -> list[ToolTypeDef]:
         tools = [self._map_tool_definition(r) for r in model_request_parameters.function_tools]
-        if model_request_parameters.result_tools:
-            tools += [self._map_tool_definition(r) for r in model_request_parameters.result_tools]
+        if model_request_parameters.output_tools:
+            tools += [self._map_tool_definition(r) for r in model_request_parameters.output_tools]
         return tools
 
     @staticmethod
@@ -185,9 +232,12 @@ class BedrockConverseModel(Model):
         messages: list[ModelMessage],
         model_settings: ModelSettings | None,
         model_request_parameters: ModelRequestParameters,
-    ) -> tuple[ModelResponse, result.Usage]:
-        response = await self._messages_create(messages, False, model_settings, model_request_parameters)
-        return await self._process_response(response)
+    ) -> ModelResponse:
+        settings = cast(BedrockModelSettings, model_settings or {})
+        response = await self._messages_create(messages, False, settings, model_request_parameters)
+        model_response = await self._process_response(response)
+        model_response.usage.requests = 1
+        return model_response
 
     @asynccontextmanager
     async def request_stream(
@@ -196,12 +246,13 @@ class BedrockConverseModel(Model):
         model_settings: ModelSettings | None,
         model_request_parameters: ModelRequestParameters,
     ) -> AsyncIterator[StreamedResponse]:
-        response = await self._messages_create(messages, True, model_settings, model_request_parameters)
+        settings = cast(BedrockModelSettings, model_settings or {})
+        response = await self._messages_create(messages, True, settings, model_request_parameters)
         yield BedrockStreamedResponse(_model_name=self.model_name, _event_stream=response)
 
-    async def _process_response(self, response: ConverseResponseTypeDef) -> tuple[ModelResponse, result.Usage]:
+    async def _process_response(self, response: ConverseResponseTypeDef) -> ModelResponse:
         items: list[ModelResponsePart] = []
-        if message := response['output'].get('message'):
+        if message := response['output'].get('message'):  # pragma: no branch
             for item in message['content']:
                 if text := item.get('text'):
                     items.append(TextPart(content=text))
@@ -215,19 +266,20 @@ class BedrockConverseModel(Model):
                             tool_call_id=tool_use['toolUseId'],
                         ),
                     )
-        usage = result.Usage(
+        u = usage.Usage(
             request_tokens=response['usage']['inputTokens'],
             response_tokens=response['usage']['outputTokens'],
             total_tokens=response['usage']['totalTokens'],
         )
-        return ModelResponse(items, model_name=self.model_name), usage
+        vendor_id = response.get('ResponseMetadata', {}).get('RequestId', None)
+        return ModelResponse(items, usage=u, model_name=self.model_name, vendor_id=vendor_id)
 
     @overload
     async def _messages_create(
         self,
         messages: list[ModelMessage],
         stream: Literal[True],
-        model_settings: ModelSettings | None,
+        model_settings: BedrockModelSettings | None,
         model_request_parameters: ModelRequestParameters,
     ) -> EventStream[ConverseStreamOutputTypeDef]:
         pass
@@ -237,7 +289,7 @@ class BedrockConverseModel(Model):
         self,
         messages: list[ModelMessage],
         stream: Literal[False],
-        model_settings: ModelSettings | None,
+        model_settings: BedrockModelSettings | None,
         model_request_parameters: ModelRequestParameters,
     ) -> ConverseResponseTypeDef:
         pass
@@ -246,33 +298,50 @@ class BedrockConverseModel(Model):
         self,
         messages: list[ModelMessage],
         stream: bool,
-        model_settings: ModelSettings | None,
+        model_settings: BedrockModelSettings | None,
         model_request_parameters: ModelRequestParameters,
     ) -> ConverseResponseTypeDef | EventStream[ConverseStreamOutputTypeDef]:
         tools = self._get_tools(model_request_parameters)
         support_tools_choice = self.model_name.startswith(('anthropic', 'us.anthropic'))
         if not tools or not support_tools_choice:
             tool_choice: ToolChoiceTypeDef = {}
-        elif not model_request_parameters.allow_text_result:
-            tool_choice = {'any': {}}
+        elif not model_request_parameters.allow_text_output:
+            tool_choice = {'any': {}}  # pragma: no cover
         else:
             tool_choice = {'auto': {}}
 
-        system_prompt, bedrock_messages = await self._map_message(messages)
+        system_prompt, bedrock_messages = await self._map_messages(messages)
         inference_config = self._map_inference_config(model_settings)
 
-        params = {
+        params: ConverseRequestTypeDef = {
             'modelId': self.model_name,
             'messages': bedrock_messages,
-            'system': [{'text': system_prompt}],
+            'system': system_prompt,
             'inferenceConfig': inference_config,
-            **(
-                {'toolConfig': {'tools': tools, **({'toolChoice': tool_choice} if tool_choice else {})}}
-                if tools
-                else {}
-            ),
         }
 
+        # Bedrock supports a set of specific extra parameters
+        if model_settings:
+            if guardrail_config := model_settings.get('bedrock_guardrail_config', None):
+                params['guardrailConfig'] = guardrail_config
+            if performance_configuration := model_settings.get('bedrock_performance_configuration', None):
+                params['performanceConfig'] = performance_configuration
+            if request_metadata := model_settings.get('bedrock_request_metadata', None):
+                params['requestMetadata'] = request_metadata
+            if additional_model_response_fields_paths := model_settings.get(
+                'bedrock_additional_model_response_fields_paths', None
+            ):
+                params['additionalModelResponseFieldPaths'] = additional_model_response_fields_paths
+            if additional_model_requests_fields := model_settings.get('bedrock_additional_model_requests_fields', None):
+                params['additionalModelRequestFields'] = additional_model_requests_fields
+            if prompt_variables := model_settings.get('bedrock_prompt_variables', None):
+                params['promptVariables'] = prompt_variables
+
+        if tools:
+            params['toolConfig'] = {'tools': tools}
+            if tool_choice:
+                params['toolConfig']['toolChoice'] = tool_choice
+
         if stream:
             model_response = await anyio.to_thread.run_sync(functools.partial(self.client.converse_stream, **params))
             model_response = model_response['stream']
@@ -289,27 +358,32 @@ class BedrockConverseModel(Model):
 
         if max_tokens := model_settings.get('max_tokens'):
             inference_config['maxTokens'] = max_tokens
-        if temperature := model_settings.get('temperature'):
+        if (temperature := model_settings.get('temperature')) is not None:
             inference_config['temperature'] = temperature
         if top_p := model_settings.get('top_p'):
             inference_config['topP'] = top_p
-        # TODO(Marcelo): This is not included in model_settings yet.
-        # if stop_sequences := model_settings.get('stop_sequences'):
-        #     inference_config['stopSequences'] = stop_sequences
+        if stop_sequences := model_settings.get('stop_sequences'):
+            inference_config['stopSequences'] = stop_sequences
 
         return inference_config
 
-    async def _map_message(self, messages: list[ModelMessage]) -> tuple[str, list[MessageUnionTypeDef]]:
-        """Just maps a `pydantic_ai.Message` to the Bedrock `MessageUnionTypeDef`."""
-        system_prompt: str = ''
+    async def _map_messages(
+        self, messages: list[ModelMessage]
+    ) -> tuple[list[SystemContentBlockTypeDef], list[MessageUnionTypeDef]]:
+        """Maps a `pydantic_ai.Message` to the Bedrock `MessageUnionTypeDef`.
+
+        Groups consecutive ToolReturnPart objects into a single user message as required by Bedrock Claude/Nova models.
+        """
+        system_prompt: list[SystemContentBlockTypeDef] = []
         bedrock_messages: list[MessageUnionTypeDef] = []
-        for m in messages:
-            if isinstance(m, ModelRequest):
-                for part in m.parts:
+        document_count: Iterator[int] = count(1)
+        for message in messages:
+            if isinstance(message, ModelRequest):
+                for part in message.parts:
                     if isinstance(part, SystemPromptPart):
-                        system_prompt += part.content
+                        system_prompt.append({'text': part.content})
                     elif isinstance(part, UserPromptPart):
-                        bedrock_messages.extend(await self._map_user_prompt(part))
+                        bedrock_messages.extend(await self._map_user_prompt(part, document_count))
                     elif isinstance(part, ToolReturnPart):
                         assert part.tool_call_id is not None
                         bedrock_messages.append(
@@ -346,9 +420,9 @@ class BedrockConverseModel(Model):
                                     ],
                                 }
                             )
-            elif isinstance(m, ModelResponse):
+            elif isinstance(message, ModelResponse):
                 content: list[ContentBlockOutputTypeDef] = []
-                for item in m.parts:
+                for item in message.parts:
                     if isinstance(item, TextPart):
                         content.append({'text': item.content})
                     else:
@@ -356,32 +430,56 @@ class BedrockConverseModel(Model):
                         content.append(self._map_tool_call(item))
                 bedrock_messages.append({'role': 'assistant', 'content': content})
             else:
-                assert_never(m)
-        return system_prompt, bedrock_messages
+                assert_never(message)
+
+        # Merge together sequential user messages.
+        processed_messages: list[MessageUnionTypeDef] = []
+        last_message: dict[str, Any] | None = None
+        for current_message in bedrock_messages:
+            if (
+                last_message is not None
+                and current_message['role'] == last_message['role']
+                and current_message['role'] == 'user'
+            ):
+                # Add the new user content onto the existing user message.
+                last_content = list(last_message['content'])
+                last_content.extend(current_message['content'])
+                last_message['content'] = last_content
+                continue
+
+            # Add the entire message to the list of messages.
+            processed_messages.append(current_message)
+            last_message = cast(dict[str, Any], current_message)
+
+        if instructions := self._get_instructions(messages):
+            system_prompt.insert(0, {'text': instructions})
+
+        return system_prompt, processed_messages
 
     @staticmethod
-    async def _map_user_prompt(part: UserPromptPart) -> list[MessageUnionTypeDef]:
+    async def _map_user_prompt(part: UserPromptPart, document_count: Iterator[int]) -> list[MessageUnionTypeDef]:
         content: list[ContentBlockUnionTypeDef] = []
         if isinstance(part.content, str):
             content.append({'text': part.content})
         else:
-            document_count = 0
             for item in part.content:
                 if isinstance(item, str):
                     content.append({'text': item})
                 elif isinstance(item, BinaryContent):
                     format = item.format
                     if item.is_document:
-                        document_count += 1
-                        name = f'Document {document_count}'
+                        name = f'Document {next(document_count)}'
                         assert format in ('pdf', 'txt', 'csv', 'doc', 'docx', 'xls', 'xlsx', 'html', 'md')
                         content.append({'document': {'name': name, 'format': format, 'source': {'bytes': item.data}}})
                     elif item.is_image:
                         assert format in ('jpeg', 'png', 'gif', 'webp')
                         content.append({'image': {'format': format, 'source': {'bytes': item.data}}})
+                    elif item.is_video:
+                        assert format in ('mkv', 'mov', 'mp4', 'webm', 'flv', 'mpeg', 'mpg', 'wmv', 'three_gp')
+                        content.append({'video': {'format': format, 'source': {'bytes': item.data}}})
                     else:
                         raise NotImplementedError('Binary content is not supported yet.')
-                elif isinstance(item, (ImageUrl, DocumentUrl)):
+                elif isinstance(item, (ImageUrl, DocumentUrl, VideoUrl)):
                     response = await cached_async_http_client().get(item.url)
                     response.raise_for_status()
                     if item.kind == 'image-url':
@@ -389,11 +487,19 @@ class BedrockConverseModel(Model):
                         assert format in ('jpeg', 'png', 'gif', 'webp'), f'Unsupported image format: {format}'
                         image: ImageBlockTypeDef = {'format': format, 'source': {'bytes': response.content}}
                         content.append({'image': image})
+
                     elif item.kind == 'document-url':
-                        document_count += 1
-                        name = f'Document {document_count}'
+                        name = f'Document {next(document_count)}'
                         data = response.content
                         content.append({'document': {'name': name, 'format': item.format, 'source': {'bytes': data}}})
+
+                    elif item.kind == 'video-url':  # pragma: no branch
+                        format = item.media_type.split('/')[1]
+                        assert format in ('mkv', 'mov', 'mp4', 'webm', 'flv', 'mpeg', 'mpg', 'wmv', 'three_gp'), (
+                            f'Unsupported video format: {format}'
+                        )
+                        video: VideoBlockTypeDef = {'format': format, 'source': {'bytes': response.content}}
+                        content.append({'video': video})
                 elif isinstance(item, AudioUrl):  # pragma: no cover
                     raise NotImplementedError('Audio is not supported yet.')
                 else:
@@ -430,13 +536,13 @@ class BedrockStreamedResponse(StreamedResponse):
             if 'messageStop' in chunk:
                 continue
             if 'metadata' in chunk:
-                if 'usage' in chunk['metadata']:
+                if 'usage' in chunk['metadata']:  # pragma: no branch
                     self._usage += self._map_usage(chunk['metadata'])
                 continue
             if 'contentBlockStart' in chunk:
                 index = chunk['contentBlockStart']['contentBlockIndex']
                 start = chunk['contentBlockStart']['start']
-                if 'toolUse' in start:
+                if 'toolUse' in start:  # pragma: no branch
                     tool_use_start = start['toolUse']
                     tool_id = tool_use_start['toolUseId']
                     tool_name = tool_use_start['name']
@@ -446,7 +552,7 @@ class BedrockStreamedResponse(StreamedResponse):
                         args=None,
                         tool_call_id=tool_id,
                     )
-                    if maybe_event:
+                    if maybe_event:  # pragma: no branch
                         yield maybe_event
             if 'contentBlockDelta' in chunk:
                 index = chunk['contentBlockDelta']['contentBlockIndex']
@@ -461,7 +567,7 @@ class BedrockStreamedResponse(StreamedResponse):
                         args=tool_use.get('input'),
                         tool_call_id=tool_id,
                     )
-                    if maybe_event:
+                    if maybe_event:  # pragma: no branch
                         yield maybe_event
 
     @property
@@ -473,8 +579,8 @@ class BedrockStreamedResponse(StreamedResponse):
         """Get the model name of the response."""
         return self._model_name
 
-    def _map_usage(self, metadata: ConverseStreamMetadataEventTypeDef) -> result.Usage:
-        return result.Usage(
+    def _map_usage(self, metadata: ConverseStreamMetadataEventTypeDef) -> usage.Usage:
+        return usage.Usage(
             request_tokens=metadata['usage']['inputTokens'],
             response_tokens=metadata['usage']['outputTokens'],
             total_tokens=metadata['usage']['totalTokens'],
@@ -492,11 +598,9 @@ class _AsyncIteratorWrapper(Generic[T]):
 
     async def __anext__(self) -> T:
         try:
-            # Run the synchronous next() call in a thread pool
-            item = await anyio.to_thread.run_sync(next, self.sync_iterator)
-            return item
+            return await anyio.to_thread.run_sync(next, self.sync_iterator)
         except RuntimeError as e:
             if type(e.__cause__) is StopIteration:
                 raise StopAsyncIteration
             else:
-                raise e
+                raise e  # pragma: lax no cover
