@@ -1,13 +1,8 @@
 from __future__ import annotations
 
-from typing import TYPE_CHECKING
-
 from httpx import Timeout
 from typing_extensions import TypedDict
 
-if TYPE_CHECKING:
-    pass
-
 
 class ModelSettings(TypedDict, total=False):
     """Settings to configure an LLM.
@@ -133,6 +128,39 @@ class ModelSettings(TypedDict, total=False):
     * Groq
     """
 
+    stop_sequences: list[str]
+    """Sequences that will cause the model to stop generating.
+
+    Supported by:
+
+    * OpenAI
+    * Anthropic
+    * Bedrock
+    * Mistral
+    * Groq
+    * Cohere
+    """
+
+    extra_headers: dict[str, str]
+    """Extra headers to send to the model.
+
+    Supported by:
+
+    * OpenAI
+    * Anthropic
+    * Groq
+    """
+
+    extra_body: object
+    """Extra body to send to the model.
+
+    Supported by:
+
+    * OpenAI
+    * Anthropic
+    * Groq
+    """
+
 
 def merge_model_settings(base: ModelSettings | None, overrides: ModelSettings | None) -> ModelSettings | None:
     """Merge two sets of model settings, preferring the overrides.
