@@ -3,12 +3,13 @@ from __future__ import annotations as _annotations
 from copy import copy
 from dataclasses import dataclass
 
+from . import _utils
 from .exceptions import UsageLimitExceeded
 
 __all__ = 'Usage', 'UsageLimits'
 
 
-@dataclass
+@dataclass(repr=False)
 class Usage:
     """LLM usage associated with a request or run.
 
@@ -28,14 +29,12 @@ class Usage:
     details: dict[str, int] | None = None
     """Any extra details returned by the model."""
 
-    def incr(self, incr_usage: Usage, *, requests: int = 0) -> None:
+    def incr(self, incr_usage: Usage) -> None:
         """Increment the usage in place.
 
         Args:
             incr_usage: The usage to increment by.
-            requests: The number of requests to increment by in addition to `incr_usage.requests`.
         """
-        self.requests += requests
         for f in 'requests', 'request_tokens', 'response_tokens', 'total_tokens':
             self_value = getattr(self, f)
             other_value = getattr(incr_usage, f)
@@ -63,11 +62,17 @@ class Usage:
             'gen_ai.usage.output_tokens': self.response_tokens,
         }
         for key, value in (self.details or {}).items():
-            result[f'gen_ai.usage.details.{key}'] = value
+            result[f'gen_ai.usage.details.{key}'] = value  # pragma: no cover
         return {k: v for k, v in result.items() if v}
 
+    def has_values(self) -> bool:
+        """Whether any values are set and non-zero."""
+        return bool(self.requests or self.request_tokens or self.response_tokens or self.details)
 
-@dataclass
+    __repr__ = _utils.dataclasses_no_defaults_repr
+
+
+@dataclass(repr=False)
 class UsageLimits:
     """Limits on model usage.
 
@@ -122,3 +127,5 @@ class UsageLimits:
         total_tokens = usage.total_tokens or 0
         if self.total_tokens_limit is not None and total_tokens > self.total_tokens_limit:
             raise UsageLimitExceeded(f'Exceeded the total_tokens_limit of {self.total_tokens_limit} ({total_tokens=})')
+
+    __repr__ = _utils.dataclasses_no_defaults_repr
